import math

from superbol.bc_polynomial import calc_bolometric_correction as bc
from superbol.bc_polynomial import quadrature_sum
from superbol.constants import mbol_zeropoint


def calc_Fbol(color_value, color_err, color_type, v_magnitude,
              v_magnitude_err):
    """Calculates the bolometric flux of a Type II-P supernova.

    Args:
        color_value (float): B-V, V-I, or B-I color of the supernova in
            magnitudes (corrected for reddening and extinction from
            the host and MWG.)
        color_err (float): Uncertainty in the photometric color.
        color_type (str): String signifying which color color_value
            represents. Valid values are "BminusV" for B-V, "VminusI"
            for V-I, and "BminusI" for B-I.
        v_magnitude (float): Photometric magnitude in the V band, corrected for
            host + MWG extinction.
        v_magnitude_err (float): Uncertainty in the V band magnitude after
            correction for host + MWG extinction.

    Returns:
        tuple: A tuple containing the bolometric flux used when calculating the
        bolometric luminosity, and the uncertainty in that number.

        (Fbol, uncertainty)

        (-999, -999) if the bolometric correction calculated from the
        color_value and color_type is -999 (which means the observed
        color is outside the range of validity of the polynomial fit.)
    """
    bolometric_correction, bc_err = bc(color_value, color_err, color_type)

    if bolometric_correction == -999:
        Fbol = -999
        Fbol_uncertainty = -999
    else:
        Fbol = 10**(-0.4 *
                    (bolometric_correction + v_magnitude + mbol_zeropoint))
        Fbol_uncertainty = (0.4 * math.log(10) * Fbol *
                            math.sqrt(bc_err**2 + v_magnitude_err**2))

    return Fbol, Fbol_uncertainty


def calc_4piDsquared(distance, distance_err):
    """Calculates :math:`4\\pi D^2`, to convert flux to luminosity.

    Args:
        distance (float): The distance to the supernova in centimeters.
        distance_err (float): The uncertainty in the distance to the supernova.

    Returns:
        tuple: A tuple containing the :math:`4 \\pi D^2`, and the uncertainty of this number.

        (4piDsquared, uncertainty)
    """
    fourPiDsquared = 4.0 * math.pi * distance**2.0
    fourPiDsquared_uncertainty = 8.0 * math.pi * distance * distance_err

    return fourPiDsquared, fourPiDsquared_uncertainty


def calc_Lbol(color_value, color_err, color_type, v_magnitude, v_magnitude_err,
              distance, distance_err):
    """Calculates the bolometric luminosity of a Type II-P Supernova.

    Args:
        color_value (float): B-V, V-I, or B-I color of the supernova in
            magnitudes (corrected for reddening and extinction from
            the host and MWG.)
        color_err (float): Uncertainty in the photometric color.
        color_type (str): String signifying which color color_value
            represents. Valid values are "BminusV" for B-V, "VminusI"
            for V-I, and "BminusI" for B-I.
        v_magnitude (float): Photometric magnitude in the V band, corrected for
            host + MWG extinction.
        v_magnitude_err (float): Uncertainty in the V band magnitude after
            correction for host + MWG extinction.
        distance (float): The distance to the supernova in centimeters.
        distance_err (float): The uncertainty in the distance to the supernova.

    Returns:
        tuple: A tuple containing the bolometric luminosity in ergs per second,
        and the uncertainty in that value.

        (Lbol, uncertainty)

        (-999, -999) if the bolometric correction is -999 (which means
        the observed color value is outside the range of vaidity of the
        polynomial fit used to determine the bolometric correction.)
    """
    Fbol, Fbol_err = calc_Fbol(color_value, color_err, color_type, v_magnitude,
                               v_magnitude_err)
    fourPiDsquared, fourPiDsquared_err = calc_4piDsquared(distance,
                                                          distance_err)

    if Fbol == -999:
        Lbol = -999
        Lbol_uncertainty = -999
    else:
        Lbol = Fbol * fourPiDsquared
        Lbol_uncertainty = quadrature_sum(fourPiDsquared * Fbol_err,
                                          Fbol * fourPiDsquared_err)

    return Lbol, Lbol_uncertainty



"""Constants for use by the bolometric correction routine
"""

# Coefficients for polynomial fit to bolometric correction - color relation
coeff_BminusV = [-0.823, 5.027, -13.409, 20.133, -18.096, 9.084, -1.950]
coeff_VminusI = [-1.355, 6.262, -2.676, -22.973, 35.524, -15.340]
coeff_BminusI = [-1.096, 3.038, -2.246, -0.497, 0.7078, 0.576, -0.713,
                 0.239, -0.027]

# Ranges of validity for polynomial fits
min_BminusV = -0.2
max_BminusV = 1.65
min_VminusI = -0.1
max_VminusI = 1.0
min_BminusI = -0.4
max_BminusI = 3.0

# RMS errors in polynomial fits
rms_err_BminusV = 0.113
rms_err_VminusI = 0.109
rms_err_BminusI = 0.091

# Zeropoint for use in the calculation of bolometric magnitude
mbol_zeropoint = 11.64



import numpy as np
from astropy import units as u


def mag2flux(magnitude, uncertainty, effective_wl, flux_at_zero_mag):
    """Converts an observed magnitude in a filter band to an average flux.

    Args:
        magnitude (float):  Apparent magnitude.
        uncertainty (float): Apparent magnitude uncertainty.
        effective_wl (float): Effective wavelength of the filter.
        flux_at_zero_mag (float): Flux at zero magnitude of the filter.

    Returns:
        tuple: A tuple of two floats:

        * the flux in :math:`erg \\; s^{-1} cm^{-2} Angstrom^{-1}`
        * the flux uncertainty in :math:`erg \\; s^{-1} cm^{-2} Angstrom^{-1}`

        (flux, flux_uncertainty)
    """
    effective_wl = effective_wl * u.AA
    flux_at_zero_mag = flux_at_zero_mag * (u.erg / (u.s * u.cm**2 * u.AA))

    flux = flux_at_zero_mag * 10**(-0.4 * magnitude)
    flux_uncertainty = np.abs(flux * -0.4 * np.log(10) * uncertainty)

    return flux.value, flux_uncertainty.value



import numpy as np

from superbol.fit_blackbody import (bb_flux_integrated, bb_total_flux,
                                    dbb_flux_integrated_dT, dbb_total_flux_dT)


def integrate_fqbol(wavelengths, fluxes, flux_uncertainties):
    """Calculate the trapezoidal rule integral of the observed `fluxes`.

    The trapezoidal rule integrates the data by assuming the function is linear between observed points, and then integrates under those line segments.
    The numpy function `trapz` is used to perform the integration, but the uncertainty in the integral due to uncertainties in the observed flux is calculated by hand using standard error propagation techniques.

    Args:
        wavelengths (list): List of wavelengths at which the flux was observed.
        fluxes (list): List of observed fluxes.
        flux_uncertainties (list): List of uncertainties in each observed flux.

    Returns:
        tuple: 2-tuple of floats.

        * The value of the integral
        * The uncertainty in the integral due to uncertainties in the fluxes.

        (fqbol, fqbol_uncertainty)
    """
    fqbol = np.trapz(fluxes, wavelengths)

    quad_terms = np.array([])

    for i, uncertainty in enumerate(flux_uncertainties):
        if i == 0:
            term = 0.5 * (wavelengths[i + 1] - wavelengths[i]) * uncertainty
            quad_terms = np.append(quad_terms, term)
        elif i == len(flux_uncertainties) - 1:
            term = 0.5 * (wavelengths[i] - wavelengths[i - 1]) * uncertainty
            quad_terms = np.append(quad_terms, term)
        else:
            term = 0.5 * (
                wavelengths[i + 1] - wavelengths[i - 1]) * uncertainty
            quad_terms = np.append(quad_terms, term)
    fqbol_uncertainty = np.sqrt(np.sum(x * x for x in quad_terms))

    fqbol_uncertainty = fqbol_uncertainty

    return fqbol, fqbol_uncertainty


def ir_correction(temperature, T_err, angular_radius, rad_err, longest_wl):
    """Apply correction for unobserved flux in the IR.

    After the temperature and angular radius has been found through fitting a
    blackbody to the observed fluxes, this function takes those values and
    integrates under the fitted blackbody function from the longest observed
    wavelength out to :math:`\\lambda = \\infty`.

    Args:
        temperature (float): Best fit blackbody temperature in Kelvin
        T_err (float): Uncertainty in best fit blackbody temperature in Kelvin
        angular_radius (float): Best fit blackbody angular radius
        rad_err (float): Uncertainty in best fit blackbody angular radius
        longest_wl (float): Longest observed wavelength

    Returns:
        tuple: 2-tuple

        * (float): The IR correction in :math:`erg \\; s^{-1} cm^{-2}`
        * (float): The uncertainty in the IR correction in the same units
    """
    ir_correction = bb_total_flux(temperature,
                                  angular_radius) - bb_flux_integrated(
                                      longest_wl, temperature, angular_radius)

    T_errterm = (dbb_total_flux_dT(temperature, angular_radius) -
                 dbb_flux_integrated_dT(longest_wl, temperature,
                                        angular_radius)) * T_err
    rad_errterm = 2 * ir_correction / angular_radius * rad_err

    ir_corr_err = np.sqrt(T_errterm**2 + rad_errterm**2)

    return ir_correction, ir_corr_err


def uv_correction_blackbody(temperature, T_err, angular_radius, rad_err,
                            shortest_wl):
    """Apply correction for unobserved flux in the UV using the blackbody fit.

    After the temperature and angular radius have been found through fitting a
    blackbody to the observed fluxes, this function takes those values and
    integrates under the fitted blackbody from the shortest observed wavelength
    down to :math:`\\lambda = 0`.

    Args:
        temperature (float): Best fit blackbody temperature in Kelvin
        T_err (float): Uncertainty in best fit blackbody temperature in Kelvin
        angular_radius (float): Best fit blackbody angular radius
        rad_err (float): Uncertainty in best fit blackbody angular radius
        shortest_wl (float): Shortest observed wavelength

    Returns:
        tuple: 2-tuple

        * (float): The UV correction in :math:`erg \\; s^{-1} cm^{-2}`
        * (float): The uncertainty in the UV correction in the same units
    """
    uv_correction = bb_flux_integrated(shortest_wl, temperature,
                                       angular_radius)

    T_errterm = dbb_flux_integrated_dT(shortest_wl, temperature,
                                       angular_radius) * T_err
    rad_errterm = 2 * uv_correction / angular_radius * rad_err

    uv_corr_err = np.sqrt(T_errterm**2 + rad_errterm**2)

    return uv_correction, uv_corr_err


def uv_correction_linear(shortest_wl, shortest_flux, shortest_flux_err):
    """Apply correction for unobserved flux in the UV using a linear function.

    This function integrates under a straight line from the shortest observed
    wavelength down to :math:`f(\\lambda) = 0` at :math:`\\lambda = 2000`
    Angstroms. This approximates the effects of line blanketing in the UV as in
    Bersten & Hamuy (2009).

    Args:
        shortest_wl (float): Shortest observed wavelength
        shortest_flux (float): Flux at shortest observed wavelength
        shortest_flux_err (float): Uncertainty in the shortest observed flux

    Returns:
        tuple: 2-tuple

        * (float): The UV correction in :math:`erg \\; s^{-1} cm^{-2}`
        * (float): The uncertainty in the UV correction in the same units
    """
    fluxes = [0.0, shortest_flux]
    wavelengths = [2000.0, shortest_wl]
    uv_correction = np.trapz(fluxes, wavelengths)
    uv_correction_err = 0.5 * (shortest_wl - 2000.0) * shortest_flux_err

    return uv_correction, uv_correction_err



import superbol.constants as constants
import math

def set_constants(color_type):
    """Sets the coefficients, validty range, and rms error of fit.

    Args:
        color_type (str): A string specifying the color combination. Must be
            "BminusV" for B-V, "VminusI" for V-I, or "BminusI" for B-I.

    Returns:
        tuple: A tuple containing the list of coefficients for the polynomial
        fit which correspond to the supplied color, the minimum value of
        the color for which the polynomial is valid, the maximum value
        of the color for which the polynomial is valid, and the rms
        error of the polynomial fit for the supplied color.

        ([coefficients], min, max, rms_error)

    Raises:
        TypeError: The argument given is not a string
        ValueError: The argument given is not one of the three valid
            strings.
    """
    if color_type == "BminusV":
        return constants.coeff_BminusV, constants.min_BminusV, \
               constants.max_BminusV, constants.rms_err_BminusV
    elif color_type == "VminusI":
        return constants.coeff_VminusI, constants.min_VminusI, \
               constants.max_VminusI, constants.rms_err_VminusI
    elif color_type == "BminusI":
        return constants.coeff_BminusI, constants.min_BminusI, \
               constants.max_BminusI, constants.rms_err_BminusI
    elif type(color_type) != str:
        raise TypeError("The argument given is not a string")
    else:
        raise ValueError("The argument given is not a valid color")

def valid_color(color_value, range_min, range_max):
    """Checks that the color value is within the range of validity.

    Each polynomial fit has a different range of validity. We need to
    make sure that the color we are feeding in is inside that range.

    Args:
        color_value (float): B-V, V-I, or B-I color of the supernova in
            magnitudes (corrected for reddening and extinction from the
            host and MWG.)
        range_min (float): Minumum value of the color range over which the fit
            is valid
        range_max (float): Maxumum value of the color range over which the fit
            is valid

    Returns:
        bool: True if the color value is within the valid range.
        False if the volor value is outside the valid range.
   """
    if range_min <= color_value <= range_max:
        return True
    else:
        return False

def calculate_polynomial_term(coefficient, variable, order):
    """Calculates a term in a polynomial.

    Args:
        coefficient (float): The coefficient to use in
            calculating the polynomial term.
        variable (float): Value to plug in for the variable in the polynomial
            term.
        order (int): Integer to use as the order of the polynomial term.

    Returns:
        float: The result of coefficient * variable**(order)

    Raises:
        TypeError: A non-integer order is given.
    """
    if type(order) != int:
        raise TypeError('Non-integer order in polynomial term')
    else:
        return coefficient * variable**(order)

def calculate_polynomial(coefficients, variable):
    """Calculates a polynomial.

    Args:
        coefficients (list): list of polynomial coefficients. The length
            of the list will be used as the order of the polynomial.
        variable (float): float to plug in for the variable in the polynomial.

    Returns:
        float: The result of summing the polynomial terms
        calculated from the coefficients and variable given.
    """
    polynomial = 0.0

    for order in range(len(coefficients)):
        polynomial += calculate_polynomial_term(coefficients[order],
                                                variable,
                                                order)

    return polynomial

def calculate_polynomial_derivative_term(coefficient, variable, order):
    """Calculates the derivative of the nth order term of a polynomial.

    Args:
        coefficient (float): The coefficient of the nth order term in the
            polynomial
        variable (float): float to plug in for the variable in the polynomial
        order (int): order of the nth order term in the polynomial (so, n.)

    Returns:
        float: The result of taking the derivative of the nth
        order term a polynomial,

        :math:`n \\cdot \\text{coefficient} \\cdot \\text{variable}^{n-1}`

        So, the edge case of taking the derivative of the zeroth-order
        term is taken care of, since you explicity multiply by the
        order of the polynomial (which is zero in the n = 0 case.)

    Raises:
        TypeError: A non-integer was passed as the order.
    """
    if type(order) != int:
        raise TypeError('Non-integer order in polynomial term')
    else:
        return order * coefficient * variable**(order - 1)

def calculate_polynomial_derivative(coefficients, variable):
    """Calculates the derivative of a polynomial.

    Args:
        coefficients (list): List of polynomial coefficients. The length
            of the list will be used as the order of the polynomial.
        variable (float): Value to plug in for the variable in the polynomial.

    Returns:
        float: The result of summing the derivatives of
        the polynomial terms calculated from the coefficients and
        variable given.
    """
    polynomial_derivative = 0.0

    for order in range(len(coefficients)):
        polynomial_derivative += calculate_polynomial_derivative_term(
            coefficients[order], variable, order)

    return polynomial_derivative

def quadrature_sum(x, y):
    """Calculate the quadrature sum of two variables x and y.

    Args:
        x (float): Variable to include in the quadrature sum
        y (float): Variable to include in the quadrature sum

    Returns:
        float: Result of calculating :math:`\\sqrt{x^2 + y^2}`
    """
    return math.sqrt(x**2 + y**2)

def calc_bolometric_correction_err(color_value, color_err, color_type):
    """Calculates the uncertainty in the bolometric correction.

    Two uncertainties are added in quadrature to get the total
    uncertainty in the bolometric correction. The first is uncertainty
    in the BC due to uncertainties in the measured color value (simple
    error propagation using a derivative.) The second is the RMS error
    inherent in the polynomial fit to the template data as reported in
    Bersten & Hamuy (2009.)

    Args:
        color_value (float): B-V, V-I, or B-I color of the supernova in
            magnitudes (corrected for reddening and extinction from
            the host and MWG.)
        color_err (float): Uncertainty in the photometric color.
        color_type (str): String signifying which color color_value
            represents.

    Returns:
        float: Uncertainty in the value of the bolometric correction
   """
    coefficients = set_constants(color_type)[0]
    rms_err = set_constants(color_type)[3]

    bc_derivative = calculate_polynomial_derivative(coefficients,
                                                     color_value)
    bc_polynomial_err = abs(bc_derivative) * color_err
    bolometric_correction_uncertainty = quadrature_sum(bc_polynomial_err,
                                                       rms_err)

    return bolometric_correction_uncertainty

def calc_bolometric_correction(color_value, color_err, color_type):
    """Calculates the bolometric correction, using a polynomial fit.

    Args:
        color_value (float): B-V, V-I, or B-I color of the supernova in
            magnitudes (corrected for reddening and extinction from the
            host and MWG.)
        color_err (float): Uncertainty in the photometric color.
        color_type (str): String signifying which color color_value represents.
            Valid values are "BminusV" for B-V, "VminusI" for V-I, and
            "BminusI" for B-I.

    Returns:
        tuple: A tuple containing the bolometric correction for use in
        calculating the bolometric luminosity of the supernova, and the
        uncertainty in that bolometric correction (if the color given
        is within the valid range of the polynomial fit.)

        (bolometric_correction, uncertainty)

        (-999, -999) if the color is outside the valid range.
    """
    bolometric_correction = 0.0

    coefficients, range_min, range_max, rms_err = set_constants(color_type)

    if valid_color(color_value, range_min, range_max):
        bolometric_correction = calculate_polynomial(coefficients,
                                                     color_value)
        uncertainty = calc_bolometric_correction_err(color_value,
                                                     color_err, color_type)
    else:
        bolometric_correction = -999
        uncertainty = -999

    return bolometric_correction, uncertainty



import numpy as np
from scipy.optimize import curve_fit
from astropy import units as u
from astropy import constants as const
from superbol.planck import *

def bb_flux(wavelength, temperature, angular_radius):
    """Observed flux at `wavelength` from a blackbody of `temperature` and `angular_radius` in cgs units

    Args:
        wavelength (float): Wavelength in Angstroms
        temperature (float): Temperature in Kelvin
        angular_radius (float): Angular radius :math:`(\\theta = \\frac{R}{D})`

    Returns:
        Astropy Quantity: blackbody flux in :math:`erg \\; s^{-1} cm^{-2} Anstrom^{-1}`
    """
    bb_flux = (np.pi * u.sr) * planck_function(wavelength, temperature) * (angular_radius)**2

    return bb_flux

def bb_flux_nounits(wavelength, temperature, angular_radius):
    """Identical to bb_flux, but returns the value rather than the Astropy Quantity.

    This function is needed because curve_fit() does not play nice with functions that return Astropy quantities.
    """
    flux = bb_flux(wavelength, temperature, angular_radius)

    return flux.value

def bb_flux_integrated(wavelength, temperature, angular_radius):
    """Integrate the planck function from :math:`\\lambda = 0` to :math:`\\lambda =` `wavelength`, then convert result to observed flux

    Args:
        wavelength (float): Wavelength in Angstroms
        temperature (float): Temperature in Kelvin
        angular_radius (float): Angular radius :math:`(\\theta = \\frac{R}{D})`

    Returns:
        float: Value of the integrated flux in :math:`erg \\; s^{-1} cm^{-2}`
    """
    bb_flux_integrated = (np.pi * u.sr) * planck_integral(wavelength, temperature) * (angular_radius)**2

    return bb_flux_integrated.value

def dbb_flux_integrated_dT(wavelength, temperature, angular_radius):
    """Take the derivative of the integrated planck function, then convert result to observed flux. This is used in error propagation calculations.

    Args:
        wavelength (float): Wavelength in Angstroms
        temperature (float): Temperature in Kelvin
        angular_radius (float): Angular radius :math:`(\\theta = \\frac{R}{D})`

    Returns:
        float: Derivative of the integrated blackbody flux at `wavelength` with respect to `temperature`
    """
    dbb_flux_integrated_dT = (np.pi * u.sr) * d_planck_integral_dT(wavelength, temperature) * (angular_radius)**2

    return dbb_flux_integrated_dT.value

def bb_total_flux(temperature, angular_radius):
    """Integrate the planck function from :math:`\\lambda = 0` to :math:`\\lambda = \\infty`, then convert result to observed flux.

    The calculation is done using the Stefan-Boltxmann law, rather than performing an actual integral using numerical integration. The `angular_radius` is included to convert the result to an observed flux.

    Args:
        temperature (float): Temperature in Kelvin
        angular_radius (float): Angular radius :math:`(\\theta = \\frac{R}{D})`

    Returns:
        float: The value of :math:`\\sigma \\frac{R^2}{D^2} T^4`
    """
    temperature = u.Quantity(temperature, u.K)
    bb_total_flux = const.sigma_sb * (angular_radius**2) * temperature**4
    bb_total_flux = bb_total_flux.to(u.erg / (u.s * u.cm**2))
    return bb_total_flux.value

def dbb_total_flux_dT(temperature, angular_radius):
    """Take the derivative of the total blackbody flux with respect to temperature

    The calculation takes a derivative of the Stefan-Boltzmann law with respect to temperature. The `angular_radius` is present to convert the result to an observed flux.

    Args:
        temperature (float): Temperature in Kelvin
        angular_radius (float): Angular radius :math:`(\\theta = \\frac{R}{D})`

    Returns:
        float: The value of :math:`4 \\sigma \\frac{R^2}{D^2} T^3`
    """
    temperature = u.Quantity(temperature, u.K)
    bb_total_flux = 4.0 * const.sigma_sb * (angular_radius**2) * temperature**3
    bb_total_flux = bb_total_flux.to(u.erg / (u.s * u.cm**2 * u.K))
    return bb_total_flux.value

def bb_fit_parameters(wavelengths, fluxes, flux_uncertainties):
    """Fit a blackbody to observed `wavelengths` and `fluxes`.

    The initial guesses for the `temperature` and `angular_radius` are :math:`T = 5000` K and :math:`\\theta = 1.0 \\times 10^{-10}`. These are typical for an extragalactic supernovae, but should be used with caution for any other objects.

    Args:
        wavelengths (list): List of wavelengths at which the fluxes were observed.
        fluxes (list): List of observed monochromatic fluxes.

    Returns:
        tuple: Tuple containing the best fit `temperature`, `angular_radius`, as well as perr, a 2-tuple containing the `temperature` error and the `angular_radius` error.

        (temperature, angular_radius, perr)
    """
    popt, pcov = curve_fit(bb_flux_nounits, wavelengths, fluxes, p0=[5000, 1.0e-10], sigma=flux_uncertainties, absolute_sigma=True)
    temperature = popt[0]
    angular_radius = popt[1]
    perr = np.sqrt(np.diag(pcov))

    return temperature, angular_radius, perr



from setuptools import setup

with open('README.rst') as file:
        long_description = file.read()

setup(
    name='SuperBoL',
    version='0.3.4',
    description='Supernova Bolometric Lightcurves',
    long_description = long_description,
    author='Jeremy A. Lusk',
    author_email='jeremy.lusk@gmail.com',
    url='https://github.com/JALusk/SuperBoL',
    packages=['superbol'],
    package_data={'superbol' : ['data/sn_data.h5']},
    license='MIT License',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2.7",
        "Topic :: Scientific/Engineering :: Astronomy",
        ],
    install_requires=[
        "astropy",
        "scipy>=0.14",
        "numpy",
        "extinction",
        "tables"
    ],
    )   



import numpy as np
from astropy import constants as const
from astropy import units as u


def planck_function(wavelength, temperature):
    """Planck function at given `wavelength` and `temperature` in cgs.

    The specific intensity of radiation from a blackbody source is given by:

    :math:`B_{\\lambda}(\\lambda, T) = \\displaystyle\\frac{C_1}{\\lambda^5} \\frac{1}{e^{\\frac{C_2}{\\lambda T}} - 1}`

    where :math:`C_1 = 2hc^2` and :math:`C_2 = hc/k_B` are the first and second radiation constants.

    Args:
        wavelength (float): Wavelength in Angstrom
        tempereature (float): Temperature in Kelvin

    Returns:
        Astropy Quantity: The specific intensity of the
        Planck function in :math:`erg \\; s^{-1} cm^{-2} sterad^{-1} Angstrom^{-1}`
    """
    wavelength = u.Quantity(wavelength, unit=u.Angstrom)
    temperature = u.Quantity(temperature, unit=u.K)

    C1 = 2.0 * const.h.cgs * const.c.cgs**2
    C2 = const.h.cgs * const.c.cgs / const.k_B.cgs

    B_lambda = (C1 / wavelength**5) / \
               (np.expm1(C2 / (wavelength * temperature))) / u.sr
    B_lambda = B_lambda.to(u.erg / (u.s * u.cm**2 * u.AA * u.sr))

    return B_lambda


def planck_integral(wavelength, temperature):
    """Integrate the Planck function over a finite wavelength interval.

    The integral is taken from :math:`\\lambda = 0` to :math:`\\lambda =` `wavelength`
    using the infinite series approximation of the integral:

    :math:`\\displaystyle\\int_{0}^{\\lambda_1} B_{\\lambda}(\\lambda, T) \\; d\\lambda = \\frac{C_1 T^4}{C_2^4} \\sum_{n = 1}^{\\infty}\\left(\\frac{x_1^3}{n} + \\frac{3x_1^2}{n^2} + \\frac{6x_1}{n^3} + \\frac{6}{n^4}\\right) e^{-nx_1}`

    where :math:`C_1 = 2hc^2` and :math:`C_2 = hc/k_B` are the first and second radiation constants, and :math:`x_1 = C_2/\\lambda_1 T`
    is a change of variables to make the integral easier to represent in this series form.

    Args:
        wavelength (float): Upper bound for the wavelength in Angstrom.
        temperature (float): Temperature in Kelvin.

    Returns:
        Astropy Quantity: Integral of the planck function in
        :math:`erg \\; s^{-1} cm^{-2} sterad^{-1}`
    """
    wavelength = u.Quantity(wavelength, unit=u.Angstrom)
    temperature = u.Quantity(temperature, unit=u.K)

    C1 = 2.0 * const.h.cgs * const.c.cgs**2
    C2 = const.h.cgs * const.c.cgs / const.k_B.cgs

    x = C2 / (wavelength.to(u.cm) * temperature)
    iterations = min(int(2 + 20 / x.value), 512)

    series = 0.0
    for i in range(1, iterations):
        term = (x**3 / i + 3 * x**2 / i**2 + 6 * x / i**3 + 6 / i**4) * np.exp(
            -i * x)
        series += term

    B_integral = (C1 * temperature**4 / C2**4) * series / u.sr

    return B_integral.to(u.erg / (u.s * u.cm**2 * u.sr))


def d_planck_integral_dT(wavelength, temperature):
    """Derivative of the integrated Planck function from :math:`\\lambda = 0` to
    :math:`\\lambda =` `wavelength` using the infinite series approximation of the
    integral. This is used in the error propagation calculation.

    Args:
        wavelength (float): Upper bound for the wavelength in Angstrom.
        temperature (float): Temperature in Kelvin.

    Returns:
        Astropy Quantity: Derivative of the integral of the planck function
        with respect to T in :math:`erg \\; s^{-1} cm^{-2} sterad^{-1} K^{-1}`
    """
    wavelength = u.Quantity(wavelength, unit=u.Angstrom)
    temperature = u.Quantity(temperature, unit=u.K)

    C1 = 2.0 * const.h.cgs * const.c.cgs**2
    C2 = const.h.cgs * const.c.cgs / const.k_B.cgs

    x = C2 / (wavelength.to(u.cm) * temperature)
    iterations = min(int(2 + 20 / x.value), 512)

    series = 0.0
    for i in range(1, iterations):
        term1 = C1 / (temperature * wavelength**4)
        term2 = 4 * C1 / (i * C2 * wavelength**3)
        term3 = 12 * C1 * temperature / (i**2 * C2**2 * wavelength**2)
        term4 = 24 * C1 * temperature**2 / (i**3 * C2**3 * wavelength)
        term5 = 24 * C1 * temperature**3 / (i**4 * C2**4)
        series += (term1 + term2 + term3 + term4 + term5) * np.exp(-i * x)

    dB_integral_dT = series / u.sr

    return dB_integral_dT.to(u.erg / (u.s * u.cm**2 * u.K * u.sr))



import numpy as np
import tables as tb
from astropy import units as u
from pkg_resources import resource_filename
import extinction

from superbol.fit_blackbody import bb_fit_parameters, bb_flux_nounits
from superbol.luminosity import calc_Lbol
from superbol.fbol import integrate_fqbol as fqbol_trapezoidal
from superbol.fbol import (ir_correction, uv_correction_blackbody,
                           uv_correction_linear)
from superbol.mag2flux import mag2flux


class SN(object):
    """A supernova is the explosion that ends the life of a star

    The SN needs to be conatained within the HDF5 database before it is used
    by SNoBoL. Once there, simply create a supernova by calling the constructor
    with the name of the SN as a string of the form "sn[YEAR][Letter(s)]"

    Attributes:
        name (str): Name of the supernova, "sn" followed by the year of first
            observation along with a letter designating the order of observation
            in that year. "sn1987a" was the first SN observed in 1987.
            "sn2000cb" was the eightieth SN observed in 2000.

    Examples:
        An example which calculates the quasi-bolometric luminosity using
        trapezoidal integration:

        >>> my_supernova = SN('sn1998a')
        >>> my_supernova.lqbol()

        In order to coorect for unobserved flux in the UV and IR, using the
        method of Bersten & Hamuy (2009), do the following:

        >>> my_supernova.lbol_direct_bh09()

        Finally, to calculate the bolometric luminosity using bolometric
        corrections based on two-filter colors as in Bersten & Hamuy (2009):

        >>> my_supernova.lbol_bc_bh09(filter1, filter2)

        where `filter1` and `filter2` are strings designating the filter to use.
        The acceptable filter combinations at this time are limited to

        =====  =========  =========
        Color  `filter1`  `filter2`
        =====  =========  =========
        B-V    "B"        "V"
        V-I    "V"        "I"
        B-I    "B"        "I"
        =====  =========  =========
    """

    def __init__(self, name):
        """Initializes the SN with supplied value for [name]"""
        self.name = name
        self.min_num_obs = 4

        self.read_hdf5()

    def read_hdf5(self):
        """Reads the hdf5 file and returns data on supernova matching [name]"""
        path_to_data = resource_filename('superbol', 'data/sn_data.h5')
        h5file = tb.open_file(path_to_data, 'r')

        self.filter_table = h5file.root.filters

        self.sn_node = h5file.get_node('/sn', self.name)

        self.phot_table = self.sn_node.phot
        self.parameter_table = self.sn_node.parameters

    def lbol_direct_bh09(self):
        """Calculate the bolometric lightcurve using the direct integration
        method published in Bersten & Hamuy 2009 (2009ApJ...701..200B)
        """
        self.convert_magnitudes_to_fluxes()
        self.deredden_fluxes()
        self.get_lbol_epochs()
        self.distance_cm, self.distance_cm_err = self.get_distance_cm()

        self.lc = np.array([[0.0, 0.0, 0.0, 0.0, 0.0]])

        for jd in self.lbol_epochs:
            names = np.array([
                x['name'] for x in self.converted_obs
                if x['jd'] == jd and x['name'] != 'z'
            ])
            wavelengths = np.array([
                x['wavelength'] for x in self.converted_obs
                if x['jd'] == jd and x['name'] != 'z'
            ])
            fluxes = np.array([
                x['flux'] for x in self.converted_obs
                if x['jd'] == jd and x['name'] != 'z'
            ])
            flux_errs = np.array([
                x['uncertainty'] for x in self.converted_obs
                if x['jd'] == jd and x['name'] != 'z'
            ])

            sort_indices = np.argsort(wavelengths)
            wavelengths = wavelengths[sort_indices]
            fluxes = fluxes[sort_indices]
            flux_errs = flux_errs[sort_indices]

            fqbol, fqbol_err = fqbol_trapezoidal(wavelengths, fluxes,
                                                 flux_errs)
            temperature, angular_radius, perr = bb_fit_parameters(
                wavelengths, fluxes, flux_errs)

            temperature_err = perr[0]
            angular_radius_err = perr[1]

            shortest_wl = np.amin(wavelengths)
            shortest_flux = np.amin(fluxes)
            shortest_flux_err = np.amin(flux_errs)
            longest_wl = np.amax(wavelengths)

            ir_corr, ir_corr_err = ir_correction(
                temperature, temperature_err, angular_radius,
                angular_radius_err, longest_wl)
            if 'U' in names:
                idx = np.nonzero(names == 'U')[0][0]
                U_flux = fluxes[idx]
                U_wl = wavelengths[idx]
                if U_flux < bb_flux_nounits(U_wl, temperature, angular_radius):
                    uv_corr, uv_corr_err = uv_correction_linear(
                        shortest_wl, shortest_flux, shortest_flux_err)
                else:
                    uv_corr, uv_corr_err = uv_correction_blackbody(
                        temperature, temperature_err, angular_radius,
                        angular_radius_err, shortest_wl)
            else:
                uv_corr, uv_corr_err = uv_correction_blackbody(
                    temperature, temperature_err, angular_radius,
                    angular_radius_err, shortest_wl)

            fbol = fqbol + ir_corr + uv_corr
            fbol_err = np.sqrt(
                np.sum(x * x for x in [fqbol_err, ir_corr_err, uv_corr_err]))
            lum = fbol * 4.0 * np.pi * self.distance_cm**2.0
            lum_err = np.sqrt((4.0 * np.pi * self.distance_cm**2 * fbol_err)**2
                              + (8.0 * np.pi * fbol * self.distance_cm *
                                 self.distance_cm_err)**2)
            phase = jd - self.parameter_table.cols.explosion_JD[0]
            phase_err = self.parameter_table.cols.explosion_JD_err[0]
            self.lc = np.append(
                self.lc, [[jd, phase, phase_err, lum, lum_err]], axis=0)

        self.lc = np.delete(self.lc, (0), axis=0)

        self.write_lbol_plaintext(self.lc, 'direct')

    def lqbol(self):
        """Calculate the quasi-bolometric lightcurve using direct integration
        with trapezoidal integration of the fluxes
        """
        self.convert_magnitudes_to_fluxes()
        self.deredden_fluxes()
        self.get_lbol_epochs()
        self.distance_cm, self.distance_cm_err = self.get_distance_cm()

        self.qbol_lc = np.array([[0.0, 0.0, 0.0, 0.0, 0.0]])

        for jd in self.lbol_epochs:
            wavelengths = np.array([
                x['wavelength'] for x in self.converted_obs if x['jd'] == jd
            ])
            fluxes = np.array(
                [x['flux'] for x in self.converted_obs if x['jd'] == jd])
            flux_errs = np.array([
                x['uncertainty'] for x in self.converted_obs if x['jd'] == jd
            ])
            names = np.array(
                [x['name'] for x in self.converted_obs if x['jd'] == jd])

            sort_indices = np.argsort(wavelengths)
            wavelengths = wavelengths[sort_indices]
            fluxes = fluxes[sort_indices]
            flux_errs = flux_errs[sort_indices]
            names = names[sort_indices]

            fqbol, fqbol_err = fqbol_trapezoidal(wavelengths, fluxes,
                                                 flux_errs)

            lqbol = fqbol * 4.0 * np.pi * self.distance_cm**2.0
            lqbol_err = np.sqrt((4.0 * np.pi * self.distance_cm**2 * fqbol_err)
                                **2 + (8.0 * np.pi * fqbol * self.distance_cm *
                                       self.distance_cm_err)**2)
            phase = jd - self.parameter_table.cols.explosion_JD[0]
            phase_err = self.parameter_table.cols.explosion_JD_err[0]
            # Quick and dirty fix for IR-only nights (don't want those in qbol calc)
            if min(wavelengths) < 10000.0:
                self.qbol_lc = np.append(
                    self.qbol_lc, [[jd, phase, phase_err, lqbol, lqbol_err]],
                    axis=0)
        self.qbol_lc = np.delete(self.qbol_lc, (0), axis=0)
        self.write_lbol_plaintext(self.qbol_lc, 'qbol')

    def lbol_bc_bh09(self, filter1, filter2):
        """Calculate the bolometric lightcurve using the bolometric corrections
        found in Bersten & Hamuy 2009 (2009ApJ...701..200B). These require
        specifying a color, taken to be filter1 - filter2"""
        self.get_magnitudes()
        self.deredden_UBVRI_magnitudes()
        self.get_bc_epochs(filter1, filter2)
        self.distance_cm, self.distance_cm_err = self.get_distance_cm()

        self.bc_lc = np.array([[0.0, 0.0, 0.0, 0.0, 0.0]])

        for i in range(len(self.bc_epochs)):
            jd = self.bc_epochs[i]
            color = self.get_bc_color(jd, filter1, filter2)
            color_err = self.get_bc_color_uncertainty(jd, filter1, filter2)
            v_mag = np.array([
                x['magnitude'] for x in self.photometry
                if x['jd'] == jd and x['name'] == 'V'
            ])
            v_mag_err = np.array([
                x['uncertainty'] for x in self.photometry
                if x['jd'] == jd and x['name'] == 'V'
            ])
            lbol_bc, lbol_bc_err = calc_Lbol(
                color, color_err, filter1 + "minus" + filter2, v_mag,
                v_mag_err, self.distance_cm, self.distance_cm_err)
            phase = jd - self.parameter_table.cols.explosion_JD[0]
            phase_err = self.parameter_table.cols.explosion_JD_err[0]
            self.bc_lc = np.append(
                self.bc_lc, [[jd, phase, phase_err, lbol_bc, lbol_bc_err]],
                axis=0)

        self.bc_lc = np.delete(self.bc_lc, (0), axis=0)
        self.write_lbol_plaintext(self.bc_lc, 'bc_' + filter1 + '-' + filter2)

    def get_bc_color(self, jd, filter1, filter2):
        """Make an array of `filter1` - `filter2` on each of the bc_epochs

        Args:
            jd (float): Julian Date of the observation
            filter1 (str): Sring designation for filter 1 ("B", for example)
            filter2 (str): String designation for filter 2 ("V", for example)

        Returns:
            float: Magnitude of filter 1 minus the magnitude of filter 2.
        """

        f1_mag = np.array([
            x['magnitude'] for x in self.photometry
            if x['jd'] == jd and x['name'] == filter1
        ])
        f2_mag = np.array([
            x['magnitude'] for x in self.photometry
            if x['jd'] == jd and x['name'] == filter2
        ])

        return f1_mag - f2_mag

    def get_bc_color_uncertainty(self, jd, filter1, filter2):
        """Make an array of :math:`\\sqrt{(\\delta \\text{filter1})^2 - (\\delta
        \\text{filter2})^2}` on each of the bc_epochs

        Args:
            jd (float): Julian Date of the observation
            filter1 (str): Sring designation for filter 1 ("B", for example)
            filter2 (str): String designation for filter 2 ("V", for example)

        Returns:
            float: Quadrature sum of the uncertainties in the magnitudes of
            filter 1 and filter 2.
        """

        f1_err = np.array([
            x['uncertainty'] for x in self.photometry
            if x['jd'] == jd and x['name'] == filter1
        ])
        f2_err = np.array([
            x['uncertainty'] for x in self.photometry
            if x['jd'] == jd and x['name'] == filter2
        ])

        return np.sqrt(f1_err**2 + f2_err**2)

    def get_magnitudes(self):
        """Build a numpy array of [`jd`, `name`, `magnitude`, `uncertainty`]
        from the data contained within the HDF5 file.
        """
        dtype = [('jd', '>f8'), ('name', 'S1'), ('magnitude', '>f8'), (
            'uncertainty', '>f8')]
        self.photometry = np.array([(0.0, '0.0', 0.0, 0.0)], dtype=dtype)

        for obs in self.phot_table.iterrows():
            filterid = obs['filter_id']
            for filt in self.filter_table.where('(filter_id == filterid)'):
                self.photometry = np.append(
                    self.photometry,
                    np.array(
                        [(obs['jd'], filt['name'], obs['magnitude'],
                          obs['uncertainty'])],
                        dtype=dtype))

        self.photometry = np.delete(self.photometry, (0), axis=0)

    def deredden_UBVRI_magnitudes(self):
        """Apply the corrections from CCM89 (1989ApJ...345..245C), Table 3 to
        the observed photometric magnitudes.

        IMPORTANT: This will only deredden the UBVRI magnitudes at the moment"""
        self.Av_gal = self.parameter_table.cols.Av_gal[0]
        self.Av_host = self.parameter_table.cols.Av_host[0]
        self.Av_tot = self.Av_gal + self.Av_host

        ccm89_corr = {'U': 1.569, 'B': 1.337, 'V': 1.0, 'R': 0.751, 'I': 0.479}

        for obs in self.photometry:
            if obs['name'] in ccm89_corr:
                obs['magnitude'] = obs['magnitude'] - ccm89_corr[obs[
                    'name']] * self.Av_tot

    def get_bc_epochs(self, filter1, filter2):
        """Get epochs for which observations of both filter1 and filter2 exist"""
        self.bc_epochs = np.array([])

        for jd_unique in np.unique(self.photometry['jd']):
            has_filter1 = False
            has_filter2 = False
            for obs in self.photometry:
                if obs['jd'] == jd_unique:
                    if obs['name'] == filter1:
                        has_filter1 = True
                    elif obs['name'] == filter2:
                        has_filter2 = True
            if has_filter1 and has_filter2:
                self.bc_epochs = np.append(self.bc_epochs, jd_unique)

    def get_distance_cm(self):
        """Get the distance to the supernova in centimeters from the HDF5 file.

        Returns:
            tuple: 2-tuple

            * (float) distance to the supernova in cm
            * (float) uncertainty in the distance to the supernova in cm
        """
        mpc_to_cm = 3.08567758E24
        distance_cm = self.parameter_table.cols.distance_Mpc[0] * mpc_to_cm
        distance_cm_err = self.parameter_table.cols.distance_Mpc_err[
            0] * mpc_to_cm
        return distance_cm, distance_cm_err

    def get_lbol_epochs(self):
        """Get only epochs with enough photometric data to calculate Lbol

        The minimum number of filters needed to calculate a luminosity is set in
        the __init__ mehod.
        """
        self.lbol_epochs = np.array([])

        for jd_unique in np.unique(self.converted_obs['jd']):
            num_obs = 0
            for obs in self.converted_obs:
                if obs['jd'] == jd_unique:
                    num_obs += 1
            if num_obs >= self.min_num_obs:
                self.lbol_epochs = np.append(self.lbol_epochs, jd_unique)

    def convert_magnitudes_to_fluxes(self):
        """Perform the magnitude to flux conversion.

        Creates an array of [`jd`, `name`, `wavelength`, `flux`, `uncertainty`]
        """
        dtype = [('jd', '>f8'), ('name', 'S1'), ('wavelength', '>f8'),
                 ('flux', '>f8'), ('uncertainty', '>f8')]
        self.converted_obs = np.array(
            [(0.0, '0.0', 0.0, 0.0, 0.0)], dtype=dtype)

        for obs in self.phot_table.iterrows():
            filterid = obs['filter_id']
            for filt in self.filter_table.where('(filter_id == filterid)'):
                flux, flux_err = mag2flux(obs['magnitude'], obs['uncertainty'],
                                          filt['eff_wl'],
                                          filt['flux_zeropoint'])
                if 909.09 <= filt['eff_wl'] <= 33333.33:
                    self.converted_obs = np.append(
                        self.converted_obs,
                        np.array(
                            [(obs['jd'], filt['name'], filt['eff_wl'], flux,
                              flux_err)],
                            dtype=dtype))

        self.converted_obs = np.delete(self.converted_obs, (0), axis=0)

    def deredden_fluxes(self):
        """Deredden the observed fluxes using the ccm89 model

        The dereddening procedure is handled by the ``apply`` method
        from the extinction package.
        """
        self.Av_gal = self.parameter_table.cols.Av_gal[0]
        self.Av_host = self.parameter_table.cols.Av_host[0]
        self.Av_tot = self.Av_gal + self.Av_host

        for obs in self.converted_obs:
            wavelength = np.array([float(obs['wavelength'])])
            obs_flux = np.array([float(obs['flux'])])
            A_lam = extinction.ccm89(wavelength, self.Av_tot, 3.1)
            obs['flux'] = extinction.apply(-A_lam, obs_flux)

    def write_lbol_plaintext(self, lightcurve, suffix):
        """Write the lightcurve to a file. Append suffix to filename"""
        filename = "lbol_" + self.name + "_" + suffix + ".dat"
        lc_file = open(filename, 'wb')
        np.savetxt(
            lc_file,
            lightcurve,
            header=self.name +
            ": JD, Phase (days), Phase err (days), Lbol (erg/s), Lbol err (erg/s)"
        )
        lc_file.close()
